@@ -38,8 +38,6 @@ struct _AtspiDeviceX11Private
   Window focused_window;
   GSource *source;
   int xi_opcode;
-  int device_id;
-  int device_id_alt;
   GSList *modifiers;
   GSList *key_grabs;
   guint virtual_mods_enabled;
@@ -175,13 +173,14 @@ grab_has_active_duplicate (AtspiDeviceX11 *x11_device, AtspiX11KeyGrab *grab)
   return FALSE;
 }
 
-static void
+static gboolean
 grab_key_aux (AtspiDeviceX11 *x11_device, Window window, int keycode, int modmask)
 {
   AtspiDeviceX11Private *priv = atspi_device_x11_get_instance_private (x11_device);
   XIGrabModifiers xi_modifiers;
   XIEventMask eventmask;
   unsigned char mask[XIMaskLen (XI_LASTEVENT)] = { 0 };
+  int ret;
 
   xi_modifiers.modifiers = modmask;
   xi_modifiers.status = 0;
@@ -193,36 +192,44 @@ grab_key_aux (AtspiDeviceX11 *x11_device, Window window, int keycode, int modmas
   XISetMask (mask, XI_KeyPress);
   XISetMask (mask, XI_KeyRelease);
 
-  XIGrabKeycode (priv->display, XIAllMasterDevices, keycode, window, XIGrabModeSync, XIGrabModeAsync, False, &eventmask, 1, &xi_modifiers);
+  ret = XIGrabKeycode (priv->display, XIAllMasterDevices, keycode, window, XIGrabModeSync, XIGrabModeAsync, False, &eventmask, 1, &xi_modifiers);
+  return (ret == 0);
 }
 
-static void
+static gboolean
 grab_key (AtspiDeviceX11 *x11_device, Window window, int keycode, int modmask)
 {
   AtspiDeviceX11Private *priv = atspi_device_x11_get_instance_private (x11_device);
+  gboolean include_numlock = !_atspi_key_is_on_keypad (keycode);
+  gboolean ret = FALSE;
 
-  grab_key_aux (x11_device, window, keycode, modmask);
+  ret |= grab_key_aux (x11_device, window, keycode, modmask);
   if (!(modmask & LockMask))
-    grab_key_aux (x11_device, window, keycode, modmask | LockMask);
-  if (!(modmask & priv->numlock_physical_mask))
+    ret |= grab_key_aux (x11_device, window, keycode, modmask | LockMask);
+  if (include_numlock && !(modmask & priv->numlock_physical_mask))
     {
-      grab_key_aux (x11_device, window, keycode, modmask | priv->numlock_physical_mask);
+      ret |= grab_key_aux (x11_device, window, keycode, modmask | priv->numlock_physical_mask);
       if (!(modmask & LockMask))
-        grab_key_aux (x11_device, window, keycode, modmask | LockMask | priv->numlock_physical_mask);
+        ret |= grab_key_aux (x11_device, window, keycode, modmask | LockMask | priv->numlock_physical_mask);
     }
+  return ret;
 }
 
-static void
+static gboolean
 enable_key_grab (AtspiDeviceX11 *x11_device, AtspiX11KeyGrab *grab)
 {
   AtspiDeviceX11Private *priv = atspi_device_x11_get_instance_private (x11_device);
+  gboolean ret;
 
-  g_return_if_fail (priv->display != NULL);
+  g_return_val_if_fail (priv->display != NULL, FALSE);
 
-  if (!grab_has_active_duplicate (x11_device, grab))
-    grab_key (x11_device, priv->focused_window, grab->kd->keycode, grab->kd->modifiers & ~ATSPI_VIRTUAL_MODIFIER_MASK);
+  if (grab_has_active_duplicate (x11_device, grab))
+    ret = TRUE;
+  else
+    ret = grab_key (x11_device, priv->focused_window, grab->kd->keycode, grab->kd->modifiers & ~ATSPI_VIRTUAL_MODIFIER_MASK);
   grab->enabled = TRUE;
   grab->window = priv->focused_window;
+  return ret;
 }
 
 static void
@@ -234,7 +241,7 @@ ungrab_key_aux (AtspiDeviceX11 *x11_device, Window window, int keycode, int modm
   xi_modifiers.modifiers = modmask;
   xi_modifiers.status = 0;
 
-  XIUngrabKeycode (priv->display, XIAllMasterDevices, keycode, window, sizeof (xi_modifiers), &xi_modifiers);
+  XIUngrabKeycode (priv->display, XIAllMasterDevices, keycode, window, 1, &xi_modifiers);
 }
 
 static void
@@ -283,9 +290,14 @@ refresh_key_grabs (AtspiDeviceX11 *x11_device)
   for (l = priv->key_grabs; l; l = l->next)
     {
       AtspiX11KeyGrab *grab = l->data;
-      gboolean new_enabled = grab_should_be_enabled (x11_device, grab);
       if (grab->window != priv->focused_window)
         disable_key_grab (x11_device, grab);
+    }
+
+  for (l = priv->key_grabs; l; l = l->next)
+    {
+      AtspiX11KeyGrab *grab = l->data;
+      gboolean new_enabled = grab_should_be_enabled (x11_device, grab);
       if (new_enabled && !grab->enabled)
         enable_key_grab (x11_device, grab);
       else if (grab->enabled && !new_enabled)
@@ -366,19 +378,11 @@ do_event_dispatch (gpointer user_data)
                   XLookupString ((XKeyEvent *) &keyevent, text, sizeof (text), &keysym, &status);
                   if (text[0] < ' ')
                     text[0] = '\0';
-                  /* The deviceid can change. Would be nice to find a better way of
-                     handling this */
-                  if (priv->device_id && priv->device_id_alt && xiDevEv->deviceid != priv->device_id && xiDevEv->deviceid != priv->device_id_alt)
-                    priv->device_id = priv->device_id_alt = 0;
-                  else if (priv->device_id && !priv->device_id_alt && xiDevEv->deviceid != priv->device_id)
-                    priv->device_id_alt = xiDevEv->deviceid;
-                  if (!priv->device_id)
-                    priv->device_id = xiDevEv->deviceid;
                   set_virtual_modifier (device, xiRawEv->detail, xevent.xcookie.evtype == XI_KeyPress);
                   modifiers = keyevent.xkey.state | priv->virtual_mods_enabled;
                   if (modifiers & priv->numlock_physical_mask)
                     modifiers |= (1 << ATSPI_MODIFIER_NUMLOCK);
-                  if (xiDevEv->deviceid == priv->device_id)
+                  if (xiDevEv->deviceid == xiDevEv->sourceid)
                     atspi_device_notify_key (ATSPI_DEVICE (device), (xevent.xcookie.evtype == XI_KeyPress), xiRawEv->detail, keysym, modifiers, text);
                   /* otherwise it's probably a duplicate event from a key grab */
                   XFreeEventData (priv->display, &xevent.xcookie);
@@ -648,19 +652,31 @@ atspi_device_x11_finalize (GObject *object)
   device_x11_parent_class->finalize (object);
 }
 
-static void
+static gboolean
 atspi_device_x11_add_key_grab (AtspiDevice *device, AtspiKeyDefinition *kd)
 {
   AtspiDeviceX11 *x11_device = ATSPI_DEVICE_X11 (device);
   AtspiDeviceX11Private *priv = atspi_device_x11_get_instance_private (x11_device);
   AtspiX11KeyGrab *grab;
+  gboolean ret;
 
   grab = g_new0 (AtspiX11KeyGrab, 1);
   grab->kd = g_boxed_copy (ATSPI_TYPE_KEY_DEFINITION, kd);
   grab->enabled = FALSE;
-  priv->key_grabs = g_slist_append (priv->key_grabs, grab);
   if (grab_should_be_enabled (x11_device, grab))
-    enable_key_grab (x11_device, grab);
+    ret = enable_key_grab (x11_device, grab);
+  else
+    ret = TRUE;
+
+  if (ret)
+    priv->key_grabs = g_slist_append (priv->key_grabs, grab);
+  else
+    {
+      g_boxed_free (ATSPI_TYPE_KEY_DEFINITION, grab->kd);
+      g_free (grab);
+    }
+
+  return ret;
 }
 
 static void
@@ -758,6 +774,35 @@ atspi_device_x11_ungrab_keyboard (AtspiDevice *device)
 #endif
 }
 
+static void
+atspi_device_x11_generate_mouse_event (AtspiDevice *device, AtspiAccessible *obj, gint x, gint y, const gchar *name, GError **error)
+{
+  AtspiPoint *p;
+
+  p = atspi_component_get_position (ATSPI_COMPONENT (obj), ATSPI_COORD_TYPE_SCREEN, error);
+  if (p->y == -1 && atspi_accessible_get_role (obj, NULL) == ATSPI_ROLE_APPLICATION)
+    {
+      g_clear_error (error);
+      g_free (p);
+      AtspiAccessible *child = atspi_accessible_get_child_at_index (obj, 0, NULL);
+      if (child)
+        {
+          p = atspi_component_get_position (ATSPI_COMPONENT (child), ATSPI_COORD_TYPE_SCREEN, error);
+          g_object_unref (child);
+        }
+    }
+
+  if (p->y == -1 || p->x == -1)
+    return;
+
+  x += p->x;
+  y += p->y;
+  g_free (p);
+
+  /* TODO: do this in process */
+  atspi_generate_mouse_event (x, y, name, error);
+}
+
 static void
 atspi_device_x11_class_init (AtspiDeviceX11Class *klass)
 {
@@ -773,6 +818,7 @@ atspi_device_x11_class_init (AtspiDeviceX11Class *klass)
   device_class->get_locked_modifiers = atspi_device_x11_get_locked_modifiers;
   device_class->grab_keyboard = atspi_device_x11_grab_keyboard;
   device_class->ungrab_keyboard = atspi_device_x11_ungrab_keyboard;
+  device_class->generate_mouse_event = atspi_device_x11_generate_mouse_event;
   object_class->finalize = atspi_device_x11_finalize;
 }
 
